@@ -53,6 +53,7 @@
 static void     notify_notification_class_init (NotifyNotificationClass *klass);
 static void     notify_notification_init       (NotifyNotification *sp);
 static void     notify_notification_finalize   (GObject            *object);
+static void     notify_notification_dispose    (GObject            *object);
 
 typedef struct
 {
@@ -70,12 +71,9 @@ struct _NotifyNotificationPrivate
         char           *body;
         char           *activation_token;
 
-        const char     *snap_path;
-        const char     *snap_name;
-        char           *snap_app;
-
         /* NULL to use icon data. Anything else to have server lookup icon */
         char           *icon_name;
+        GdkPixbuf      *icon_pixbuf;
 
         /*
          * -1   = use server default
@@ -83,6 +81,7 @@ struct _NotifyNotificationPrivate
          *  > 0 = Number of milliseconds before we timeout
          */
         gint            timeout;
+        guint           portal_timeout_id;
 
         GSList         *actions;
         GHashTable     *action_map;
@@ -154,6 +153,7 @@ notify_notification_class_init (NotifyNotificationClass *klass)
         object_class->constructor = notify_notification_constructor;
         object_class->get_property = notify_notification_get_property;
         object_class->set_property = notify_notification_set_property;
+        object_class->dispose = notify_notification_dispose;
         object_class->finalize = notify_notification_finalize;
 
         /**
@@ -238,9 +238,9 @@ notify_notification_class_init (NotifyNotificationClass *klass)
                                          g_param_spec_int ("closed-reason",
                                                            "Closed Reason",
                                                            "The reason code for why the notification was closed",
-                                                           -1,
+                                                           NOTIFY_CLOSED_REASON_UNSET,
                                                            G_MAXINT32,
-                                                           -1,
+                                                           NOTIFY_CLOSED_REASON_UNSET,
                                                            G_PARAM_READABLE
                                                            | G_PARAM_STATIC_NAME
                                                            | G_PARAM_STATIC_NICK
@@ -356,110 +356,12 @@ destroy_pair (CallbackPair *pair)
         g_free (pair);
 }
 
-static void
-maybe_initialize_snap (NotifyNotification *obj)
-{
-        NotifyNotificationPrivate *priv = obj->priv;
-        gchar *cgroup_contents = NULL;
-
-        priv->snap_path = g_getenv ("SNAP");
-        if (priv->snap_path == NULL)
-                return;
-
-        if (*priv->snap_path == '\0' ||
-            !strchr (priv->snap_path, G_DIR_SEPARATOR)) {
-                priv->snap_path = NULL;
-                return;
-        }
-
-        priv->snap_name = g_getenv ("SNAP_NAME");
-        if (priv->snap_name && *priv->snap_name == '\0') {
-                priv->snap_name = NULL;
-        }
-
-        if (g_file_get_contents ("/proc/self/cgroup", &cgroup_contents,
-                                 NULL, NULL)) {
-                gchar **lines = g_strsplit (cgroup_contents, "\n", -1);
-                gchar *found_snap_name = NULL;
-                gint i;
-
-                for (i = 0; lines[i]; ++i) {
-                        gchar **parts = g_strsplit (lines[i], ":", 3);
-                        gchar *basename;
-                        gchar **ns;
-                        guint ns_length;
-
-                        if (g_strv_length (parts) != 3) {
-                                g_strfreev (parts);
-                                continue;
-                        }
-
-                        basename = g_path_get_basename (parts[2]);
-                        g_strfreev (parts);
-
-                        if (!basename) {
-                                continue;
-                        }
-
-                        ns = g_strsplit (basename, ".", -1);
-                        ns_length = g_strv_length (ns);
-                        g_free (basename);
-
-                        if (ns_length < 2 || !g_str_equal (ns[0], "snap")) {
-                                g_strfreev (ns);
-                                continue;
-                        }
-
-                        if (priv->snap_name == NULL) {
-                                g_free (found_snap_name);
-                                found_snap_name = g_strdup (ns[1]);
-                        }
-
-                        if (ns_length < 3) {
-                                g_strfreev (ns);
-                                continue;
-                        }
-
-                        if (priv->snap_name == NULL) {
-                                priv->snap_name = found_snap_name;
-                                found_snap_name = NULL;
-                        }
-
-                        if (g_str_equal (ns[1], priv->snap_name)) {
-                                priv->snap_app = g_strdup (ns[2]);
-                                g_strfreev (ns);
-                                break;
-                        }
-
-                        g_strfreev (ns);
-                }
-
-                if (priv->snap_name == NULL && found_snap_name != NULL) {
-                        priv->snap_name = found_snap_name;
-                        found_snap_name = NULL;
-                }
-
-                g_strfreev (lines);
-                g_free (found_snap_name);
-        }
-
-        if (priv->snap_app == NULL) {
-                priv->snap_app = g_strdup (priv->snap_name);
-        }
-
-        g_debug ("SNAP path: %s", priv->snap_path);
-        g_debug ("SNAP name: %s", priv->snap_name);
-        g_debug ("SNAP app: %s", priv->snap_app);
-
-        g_free (cgroup_contents);
-}
-
 static void
 notify_notification_init (NotifyNotification *obj)
 {
         obj->priv = g_new0 (NotifyNotificationPrivate, 1);
         obj->priv->timeout = NOTIFY_EXPIRES_DEFAULT;
-        obj->priv->closed_reason = -1;
+        obj->priv->closed_reason = NOTIFY_CLOSED_REASON_UNSET;
         obj->priv->hints = g_hash_table_new_full (g_str_hash,
                                                   g_str_equal,
                                                   g_free,
@@ -469,8 +371,22 @@ notify_notification_init (NotifyNotification *obj)
                                                        g_str_equal,
                                                        g_free,
                                                        (GDestroyNotify) destroy_pair);
+}
 
-        maybe_initialize_snap (obj);
+static void
+notify_notification_dispose (GObject *object)
+{
+        NotifyNotification        *obj = NOTIFY_NOTIFICATION (object);
+        NotifyNotificationPrivate *priv = obj->priv;
+
+        if (priv->portal_timeout_id) {
+                g_source_remove (priv->portal_timeout_id);
+                priv->portal_timeout_id = 0;
+        }
+
+        g_clear_object (&priv->icon_pixbuf);
+
+        G_OBJECT_CLASS (parent_class)->dispose (object);
 }
 
 static void
@@ -487,7 +403,6 @@ notify_notification_finalize (GObject *object)
         g_free (priv->body);
         g_free (priv->icon_name);
         g_free (priv->activation_token);
-        g_free (priv->snap_app);
 
         if (priv->actions != NULL) {
                 g_slist_foreach (priv->actions, (GFunc) g_free, NULL);
@@ -510,6 +425,18 @@ notify_notification_finalize (GObject *object)
         G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
+static gboolean
+maybe_warn_portal_unsupported_feature (const char *feature_name)
+{
+        if (!_notify_uses_portal_notifications ()) {
+                return FALSE;
+        }
+
+        g_message ("%s is not available when using Portal Notifications",
+                   feature_name);
+        return TRUE;
+}
+
 /**
  * notify_notification_new:
  * @summary: The required summary text.
@@ -596,7 +523,6 @@ static gchar *
 try_prepend_snap_desktop (NotifyNotification *notification,
                           const gchar        *desktop)
 {
-        NotifyNotificationPrivate *priv = notification->priv;
         gchar *ret = NULL;
 
         /*
@@ -604,11 +530,11 @@ try_prepend_snap_desktop (NotifyNotification *notification,
          * ${SNAP_NAME}_; snap .desktop files are in the format
          * ${SNAP_NAME}_desktop_file_name
          */
-        ret = try_prepend_path (desktop, priv->snap_path);
+        ret = try_prepend_path (desktop, _notify_get_snap_path ());
 
-        if (ret == NULL && priv->snap_name != NULL &&
+        if (ret == NULL && _notify_get_snap_name () != NULL &&
             strchr (desktop, G_DIR_SEPARATOR) == NULL) {
-                ret = g_strdup_printf ("%s_%s", priv->snap_name, desktop);
+                ret = g_strdup_printf ("%s_%s", _notify_get_snap_name (), desktop);
         }
 
         return ret;
@@ -619,7 +545,7 @@ try_prepend_snap (NotifyNotification *notification,
                   const gchar        *value)
 {
         /* hardcoded paths to icons might be relocated under $SNAP */
-        return try_prepend_path (value, notification->priv->snap_path);
+        return try_prepend_path (value, _notify_get_snap_path ());
 }
 
 
@@ -698,6 +624,80 @@ notify_notification_update (NotifyNotification *notification,
         return TRUE;
 }
 
+static char *
+get_portal_notification_id (NotifyNotification *notification)
+{
+        char *app_id;
+        char *notification_id;
+
+        g_assert (_notify_uses_portal_notifications ());
+
+        if (_notify_get_snap_name ()) {
+                app_id = g_strdup_printf ("snap.%s_%s",
+                                          _notify_get_snap_name (),
+                                          _notify_get_snap_app ());
+        } else {
+                app_id = g_strdup_printf ("flatpak.%s",
+                                          _notify_get_flatpak_app ());
+        }
+
+        notification_id = g_strdup_printf ("libnotify-%s-%s-%u",
+                                           app_id,
+                                           notify_get_app_name (),
+                                           notification->priv->id);
+
+        g_free (app_id);
+
+        return notification_id;
+}
+
+static gboolean
+activate_action (NotifyNotification *notification,
+                 const gchar        *action)
+{
+        CallbackPair *pair;
+
+        pair = g_hash_table_lookup (notification->priv->action_map, action);
+
+        if (!pair) {
+                return FALSE;
+        }
+
+        /* Some clients have assumed it is safe to unref the
+         * Notification at the end of their NotifyActionCallback
+         * so we add a temporary ref until we're done with it.
+         */
+        g_object_ref (notification);
+
+        notification->priv->activating = TRUE;
+        pair->cb (notification, (char *) action, pair->user_data);
+        notification->priv->activating = FALSE;
+        g_free (notification->priv->activation_token);
+        notification->priv->activation_token = NULL;
+
+        g_object_unref (notification);
+
+        return TRUE;
+}
+
+static gboolean
+close_notification (NotifyNotification *notification,
+                    NotifyClosedReason  reason)
+{
+        if (notification->priv->closed_reason != NOTIFY_CLOSED_REASON_UNSET ||
+            reason == NOTIFY_CLOSED_REASON_UNSET) {
+                return FALSE;
+        }
+
+        g_object_ref (G_OBJECT (notification));
+        notification->priv->closed_reason = reason;
+        g_signal_emit (notification, signals[SIGNAL_CLOSED], 0);
+        notification->priv->id = 0;
+        g_object_unref (G_OBJECT (notification));
+
+        return TRUE;
+}
+
 static void
 proxy_g_signal_cb (GDBusProxy *proxy,
                    const char *sender_name,
@@ -705,8 +705,12 @@ proxy_g_signal_cb (GDBusProxy *proxy,
                    GVariant   *parameters,
                    NotifyNotification *notification)
 {
+        const char *interface;
+
         g_return_if_fail (NOTIFY_IS_NOTIFICATION (notification));
 
+        interface = g_dbus_proxy_get_interface_name (proxy);
+
         if (g_strcmp0 (signal_name, "NotificationClosed") == 0 &&
             g_variant_is_of_type (parameters, G_VARIANT_TYPE ("(uu)"))) {
                 guint32 id, reason;
@@ -715,43 +719,21 @@ proxy_g_signal_cb (GDBusProxy *proxy,
                 if (id != notification->priv->id)
                         return;
 
-                g_object_ref (G_OBJECT (notification));
-                notification->priv->closed_reason = reason;
-                g_signal_emit (notification, signals[SIGNAL_CLOSED], 0);
-                notification->priv->id = 0;
-                g_object_unref (G_OBJECT (notification));
+                close_notification (notification, reason);
         } else if (g_strcmp0 (signal_name, "ActionInvoked") == 0 &&
+                   g_str_equal (interface, NOTIFY_DBUS_CORE_INTERFACE) &&
                    g_variant_is_of_type (parameters, G_VARIANT_TYPE ("(us)"))) {
                 guint32 id;
                 const char *action;
-                CallbackPair *pair;
 
                 g_variant_get (parameters, "(u&s)", &id, &action);
 
                 if (id != notification->priv->id)
                         return;
 
-                pair = (CallbackPair *) g_hash_table_lookup (notification->priv->action_map,
-                                                            action);
-
-                if (pair == NULL) {
-                        if (g_ascii_strcasecmp (action, "default")) {
-                                g_warning ("Received unknown action %s", action);
-                        }
-                } else {
-                        /* Some clients have assumed it is safe to unref the
-                         * Notification at the end of their NotifyActionCallback
-                         * so we add a temporary ref until we're done with it.
-                         */
-                        g_object_ref (notification);
-
-                        notification->priv->activating = TRUE;
-                        pair->cb (notification, (char *) action, pair->user_data);
-                        notification->priv->activating = FALSE;
-                        g_free (notification->priv->activation_token);
-                        notification->priv->activation_token = NULL;
-
-                        g_object_unref (notification);
+                if (!activate_action (notification, action) &&
+                    g_ascii_strcasecmp (action, "default")) {
+                        g_warning ("Received unknown action %s", action);
                 }
         } else if (g_strcmp0 (signal_name, "ActivationToken") == 0 &&
                    g_variant_is_of_type (parameters, G_VARIANT_TYPE ("(us)"))) {
@@ -765,9 +747,306 @@ proxy_g_signal_cb (GDBusProxy *proxy,
 
                 g_free (notification->priv->activation_token);
                 notification->priv->activation_token = g_strdup (activation_token);
+        } else if (g_str_equal (signal_name, "ActionInvoked") &&
+                   g_str_equal (interface, NOTIFY_PORTAL_DBUS_CORE_INTERFACE) &&
+                   g_variant_is_of_type (parameters, G_VARIANT_TYPE ("(ssav)"))) {
+                char *notification_id;
+                const char *id;
+                const char *action;
+                GVariant *parameter;
+
+                g_variant_get (parameters, "(&s&s@av)", &id, &action, &parameter);
+                g_variant_unref (parameter);
+
+                notification_id = get_portal_notification_id (notification);
+
+                if (!g_str_equal (notification_id, id)) {
+                        g_free (notification_id);
+                        return;
+                }
+
+                if (!activate_action (notification, action) &&
+                    g_str_equal (action, "default-action") &&
+                    !_notify_get_snap_app ()) {
+                        g_warning ("Received unknown action %s", action);
+                }
+
+                close_notification (notification, NOTIFY_CLOSED_REASON_DISMISSED);
+
+                g_free (notification_id);
+        } else {
+                g_debug ("Unhandled signal '%s.%s'", interface, signal_name);
         }
 }
 
+static gboolean
+remove_portal_notification (GDBusProxy         *proxy,
+                            NotifyNotification *notification,
+                            NotifyClosedReason  reason,
+                            GError            **error)
+{
+        GVariant *ret;
+        gchar *notification_id;
+
+        if (notification->priv->portal_timeout_id) {
+                g_source_remove (notification->priv->portal_timeout_id);
+                notification->priv->portal_timeout_id = 0;
+        }
+
+        notification_id = get_portal_notification_id (notification);
+
+        ret = g_dbus_proxy_call_sync (proxy,
+                                      "RemoveNotification",
+                                      g_variant_new ("(s)", notification_id),
+                                      G_DBUS_CALL_FLAGS_NONE,
+                                      -1,
+                                      NULL,
+                                      error);
+
+        g_free (notification_id);
+
+        if (!ret) {
+                return FALSE;
+        }
+
+        close_notification (notification, reason);
+
+        g_variant_unref (ret);
+
+        return TRUE;
+}
+
+static gboolean
+on_portal_timeout (gpointer data)
+{
+        NotifyNotification *notification = data;
+        GDBusProxy *proxy;
+
+        notification->priv->portal_timeout_id = 0;
+
+        proxy = _notify_get_proxy (NULL);
+        if (proxy == NULL) {
+                return FALSE;
+        }
+
+        remove_portal_notification (proxy, notification,
+                                    NOTIFY_CLOSED_REASON_EXPIRED, NULL);
+        return FALSE;
+}
+
+static GIcon *
+get_notification_gicon (NotifyNotification  *notification,
+                        GError             **error)
+{
+        NotifyNotificationPrivate *priv = notification->priv;
+        GFileInputStream *input;
+        GFile *file = NULL;
+        GIcon *gicon = NULL;
+
+        if (priv->icon_pixbuf) {
+                return G_ICON (g_object_ref (priv->icon_pixbuf));
+        }
+
+        if (!priv->icon_name) {
+                return NULL;
+        }
+
+        if (strstr (priv->icon_name, "://")) {
+                file = g_file_new_for_uri (priv->icon_name);
+        } else if (g_file_test (priv->icon_name, G_FILE_TEST_EXISTS)) {
+                file = g_file_new_for_path (priv->icon_name);
+        } else {
+                gicon = g_themed_icon_new (priv->icon_name);
+        }
+
+        if (!file) {
+                return gicon;
+        }
+
+        input = g_file_read (file, NULL, error);
+
+        if (input) {
+                GByteArray *bytes_array = g_byte_array_new ();
+                guint8 buf[1024];
+
+                while (TRUE) {
+                        gssize read;
+
+                        read = g_input_stream_read (G_INPUT_STREAM (input),
+                                                    buf,
+                                                    G_N_ELEMENTS (buf),
+                                                    NULL, NULL);
+
+                        if (read > 0) {
+                                g_byte_array_append (bytes_array, buf, read);
+                        } else {
+                                if (read < 0) {
+                                        g_byte_array_unref (bytes_array);
+                                        bytes_array = NULL;
+                                }
+
+                                break;
+                        }
+                }
+
+                if (bytes_array && bytes_array->len) {
+                        GBytes *bytes;
+
+                        bytes = g_byte_array_free_to_bytes (bytes_array);
+                        bytes_array = NULL;
+
+                        gicon = g_bytes_icon_new (bytes);
+                } else if (bytes_array) {
+                        g_byte_array_unref (bytes_array);
+                }
+        }
+
+        g_clear_object (&input);
+        g_clear_object (&file);
+
+        return gicon;
+}
+
+static gboolean
+add_portal_notification (GDBusProxy         *proxy,
+                         NotifyNotification *notification,
+                         GError            **error)
+{
+        GIcon *icon;
+        GVariant *urgency;
+        GVariant *ret;
+        GVariantBuilder builder;
+        NotifyNotificationPrivate *priv = notification->priv;
+        GError *local_error = NULL;
+        static guint32 portal_notification_count = 0;
+        char *notification_id;
+
+        g_variant_builder_init (&builder, G_VARIANT_TYPE_VARDICT);
+
+        g_variant_builder_add (&builder, "{sv}", "title",
+                               g_variant_new_string (priv->summary ? priv->summary : ""));
+        g_variant_builder_add (&builder, "{sv}", "body",
+                               g_variant_new_string (priv->body ? priv->body : ""));
+
+        if (g_hash_table_lookup (priv->action_map, "default")) {
+                g_variant_builder_add (&builder, "{sv}", "default-action",
+                                       g_variant_new_string ("default"));
+        } else if (g_hash_table_lookup (priv->action_map, "DEFAULT")) {
+                g_variant_builder_add (&builder, "{sv}", "default-action",
+                                       g_variant_new_string ("DEFAULT"));
+        } else if (_notify_get_snap_app ()) {
+                /* In the snap case we may need to ensure that a default-action
+                 * is set to ensure that we will use the FDO notification daemon
+                 * and won't fallback to GTK one, as app-id won't match.
+                 * See: https://github.com/flatpak/xdg-desktop-portal/issues/769
+                 */
+                g_variant_builder_add (&builder, "{sv}", "default-action",
+                                       g_variant_new_string ("snap-fake-default-action"));
+        }
+
+        if (priv->has_nondefault_actions) {
+                GVariantBuilder buttons;
+                GSList *l;
+
+                g_variant_builder_init (&buttons, G_VARIANT_TYPE ("aa{sv}"));
+
+                for (l = priv->actions; l && l->next; l = l->next->next) {
+                        GVariantBuilder button;
+                        const char *action;
+                        const char *label;
+
+                        g_variant_builder_init (&button, G_VARIANT_TYPE_VARDICT);
+
+                        action = l->data;
+                        label = l->next->data;
+
+                        g_variant_builder_add (&button, "{sv}", "action",
+                                               g_variant_new_string (action));
+                        g_variant_builder_add (&button, "{sv}", "label",
+                                               g_variant_new_string (label));
+
+                        g_variant_builder_add (&buttons, "@a{sv}",
+                                               g_variant_builder_end (&button));
+                }
+
+                g_variant_builder_add (&builder, "{sv}", "buttons",
+                                       g_variant_builder_end (&buttons));
+        }
+
+        urgency = g_hash_table_lookup (notification->priv->hints, "urgency");
+        if (urgency) {
+                switch (g_variant_get_byte (urgency)) {
+                case NOTIFY_URGENCY_LOW:
+                        g_variant_builder_add (&builder, "{sv}", "priority",
+                                               g_variant_new_string ("low"));
+                        break;
+                case NOTIFY_URGENCY_NORMAL:
+                        g_variant_builder_add (&builder, "{sv}", "priority",
+                                               g_variant_new_string ("normal"));
+                        break;
+                case NOTIFY_URGENCY_CRITICAL:
+                        g_variant_builder_add (&builder, "{sv}", "priority",
+                                               g_variant_new_string ("urgent"));
+                        break;
+                default:
+                        g_warn_if_reached ();
+                }
+        }
+
+        icon = get_notification_gicon (notification, &local_error);
+        if (icon) {
+                GVariant *serialized_icon = g_icon_serialize (icon);
+
+                g_variant_builder_add (&builder, "{sv}", "icon",
+                                       serialized_icon);
+                g_variant_unref (serialized_icon);
+                g_clear_object (&icon);
+        } else if (local_error) {
+                g_propagate_error (error, local_error);
+                return FALSE;
+        }
+
+        if (!priv->id) {
+                priv->id = ++portal_notification_count;
+        } else if (priv->closed_reason == NOTIFY_CLOSED_REASON_UNSET) {
+                remove_portal_notification (proxy, notification,
+                                            NOTIFY_CLOSED_REASON_UNSET, NULL);
+        }
+
+        notification_id = get_portal_notification_id (notification);
+
+        ret = g_dbus_proxy_call_sync (proxy,
+                                      "AddNotification",
+                                      g_variant_new ("(s@a{sv})",
+                                                     notification_id,
+                                                     g_variant_builder_end (&builder)),
+                                      G_DBUS_CALL_FLAGS_NONE,
+                                      -1,
+                                      NULL,
+                                      error);
+
+        if (priv->portal_timeout_id) {
+                g_source_remove (priv->portal_timeout_id);
+                priv->portal_timeout_id = 0;
+        }
+
+        g_free (notification_id);
+
+        if (!ret) {
+                return FALSE;
+        }
+
+        if (priv->timeout > 0) {
+                priv->portal_timeout_id = g_timeout_add (priv->timeout,
+                                                         on_portal_timeout,
+                                                         notification);
+        }
+
+        g_variant_unref (ret);
+
+        return TRUE;
+}
+
 /**
  * notify_notification_show:
  * @notification: The notification.
@@ -789,9 +1068,7 @@ notify_notification_show (NotifyNotification *notification,
         GHashTableIter             iter;
         gpointer                   key, data;
         GVariant                  *result;
-#ifdef GLIB_VERSION_2_32
         GApplication              *application = NULL;
-#endif
 
         g_return_val_if_fail (notification != NULL, FALSE);
         g_return_val_if_fail (NOTIFY_IS_NOTIFICATION (notification), FALSE);
@@ -815,6 +1092,10 @@ notify_notification_show (NotifyNotification *notification,
                                                                notification);
         }
 
+        if (_notify_uses_portal_notifications ()) {
+                return add_portal_notification (proxy, notification, error);
+        }
+
         g_variant_builder_init (&actions_builder, G_VARIANT_TYPE ("as"));
         for (l = priv->actions; l != NULL; l = l->next) {
                 g_variant_builder_add (&actions_builder, "s", l->data);
@@ -831,13 +1112,13 @@ notify_notification_show (NotifyNotification *notification,
                                        g_variant_new_int64 (getpid ()));
         }
 
-        if (priv->snap_app &&
+        if (_notify_get_snap_app () &&
             g_hash_table_lookup (priv->hints, "desktop-entry") == NULL) {
                 gchar *snap_desktop;
 
                 snap_desktop = g_strdup_printf ("%s_%s",
-                                                priv->snap_name,
-                                                priv->snap_app);
+                                                _notify_get_snap_name (),
+                                                _notify_get_snap_app ());
 
                 g_debug ("Using desktop entry: %s", snap_desktop);
                 g_variant_builder_add (&hints_builder, "{sv}",
@@ -845,8 +1126,7 @@ notify_notification_show (NotifyNotification *notification,
                                        g_variant_new_take_string (snap_desktop));
         }
 
-#ifdef GLIB_VERSION_2_32
-        if (!priv->snap_app) {
+        if (!_notify_get_snap_app ()) {
                 application = g_application_get_default ();
         }
 
@@ -861,7 +1141,6 @@ notify_notification_show (NotifyNotification *notification,
                                        g_variant_new_string (application_id));
             }
         }
-#endif
 
         /* TODO: make this nonblocking */
         result = g_dbus_proxy_call_sync (proxy,
@@ -940,6 +1219,10 @@ notify_notification_set_category (NotifyNotification *notification,
         g_return_if_fail (notification != NULL);
         g_return_if_fail (NOTIFY_IS_NOTIFICATION (notification));
 
+        if (maybe_warn_portal_unsupported_feature ("Category")) {
+                return;
+        }
+
         if (category != NULL && category[0] != '\0') {
                 notify_notification_set_hint_string (notification,
                                                      "category",
@@ -1017,11 +1300,18 @@ notify_notification_set_image_from_pixbuf (NotifyNotification *notification,
                 hint_name = "icon_data";
         }
 
+        g_clear_object (&notification->priv->icon_pixbuf);
+
         if (pixbuf == NULL) {
                 notify_notification_set_hint (notification, hint_name, NULL);
                 return;
         }
 
+        if (_notify_uses_portal_notifications ()) {
+                notification->priv->icon_pixbuf = g_object_ref (pixbuf);
+                return;
+        }
+
         g_object_get (pixbuf,
                       "width", &width,
                       "height", &height,
@@ -1078,7 +1368,7 @@ maybe_parse_snap_hint_value (NotifyNotification *notification,
 {
         StringParserFunc parse_func = NULL;
 
-        if (!notification->priv->snap_path)
+        if (!_notify_get_snap_path ())
                 return value;
 
         if (g_strcmp0 (key, "desktop-entry") == 0) {
@@ -1145,6 +1435,10 @@ notify_notification_set_app_name (NotifyNotification *notification,
 {
         g_return_if_fail (NOTIFY_IS_NOTIFICATION (notification));
 
+        if (maybe_warn_portal_unsupported_feature ("App Name")) {
+                return;
+        }
+
         g_free (notification->priv->app_name);
         notification->priv->app_name = g_strdup (app_name);
 
@@ -1385,6 +1679,7 @@ notify_notification_add_action (NotifyNotification  *notification,
 
 /**
  * notify_notification_get_activation_token:
+ * @notification: The notification.
  *
  * If an an action is currently being activated, return the activation token.
  * This function is intended to be used in a #NotifyActionCallback to get
@@ -1399,11 +1694,9 @@ const char *
 notify_notification_get_activation_token (NotifyNotification *notification)
 {
         g_return_val_if_fail (NOTIFY_IS_NOTIFICATION (notification), NULL);
+        g_return_val_if_fail (notification->priv->activating, NULL);
 
-        if (notification->priv->activating)
-                return notification->priv->activation_token;
-
-        return NULL;
+        return notification->priv->activation_token;
 }
 
 gboolean
@@ -1442,6 +1735,12 @@ notify_notification_close (NotifyNotification *notification,
                 return FALSE;
         }
 
+        if (_notify_uses_portal_notifications ()) {
+                return remove_portal_notification (proxy, notification,
+                                                   NOTIFY_CLOSED_REASON_API_REQUEST,
+                                                   error);
+        }
+
         /* FIXME: make this nonblocking! */
         result = g_dbus_proxy_call_sync (proxy,
                                          "CloseNotification",
@@ -1466,13 +1765,17 @@ notify_notification_close (NotifyNotification *notification,
  * Returns the closed reason code for the notification. This is valid only
  * after the "closed" signal is emitted.
  *
- * Returns: The closed reason code.
+ * Since version 0.8.0 the returned value is of type #NotifyClosedReason.
+ *
+ * Returns: An integer representing the closed reason code
+ *  (Since 0.8.0 it's also a #NotifyClosedReason).
  */
 gint
 notify_notification_get_closed_reason (const NotifyNotification *notification)
 {
-        g_return_val_if_fail (notification != NULL, -1);
-        g_return_val_if_fail (NOTIFY_IS_NOTIFICATION (notification), -1);
+        g_return_val_if_fail (notification != NULL, NOTIFY_CLOSED_REASON_UNSET);
+        g_return_val_if_fail (NOTIFY_IS_NOTIFICATION (notification),
+                              NOTIFY_CLOSED_REASON_UNSET);
 
         return notification->priv->closed_reason;
 }
